# -*- coding: utf-8 -*-
# (c) 2015 Tuomas Airaksinen
#
# This file is part of Automate.
#
# Automate is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Automate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Automate.  If not, see <http://www.gnu.org/licenses/>.
#
# ------------------------------------------------------------------
#
# If you like Automate, please take a look at this page:
# http://evankelista.net/automate/

from __future__ import absolute_import
from __future__ import unicode_literals
import logging
import operator
import threading
import time
import sys

import datetime
from traits.api import (cached_property, Any, CBool, Instance, Dict, Str, CFloat,
                        List, Enum, Bool, Property, Event)
from traits.trait_errors import TraitError

from automate.common import Lock, AbstractStatusObject, CompareMixin, nomutex
from automate.worker import StatusWorkerTask, DummyStatusWorkerTask
from automate.program import ProgrammableSystemObject, DefaultProgram
from automate.systemobject import SystemObject

if sys.version_info >= (3, 0):
    TimerClass = threading.Timer
else:
    TimerClass = threading._Timer


class StatusObject(AbstractStatusObject, ProgrammableSystemObject, CompareMixin):

    """
        Baseclass for Sensors and Actuators
    """

    editable = True

    #: Determines minimum time required for switching. State change is then delayed if necessary.
    safety_delay = CFloat(0.)

    #: Determines when :attr:`.safety_delay` needs to be taken into account: when status is
    #: rising, falling or both.
    safety_mode = Enum("rising", "falling", "both")

    #: Similar to :attr:`.safety_delay`, but just delays change to make sure that events shorter
    #: than change_delay are not taken into account
    change_delay = CFloat(0.)

    #: As :attr:`.safety_mode`, but for :attr:`.change_delay`
    change_mode = Enum("rising", "falling", "both")

    #: Do not emit actuator status changes into logs
    silent = CBool(False)

    #: Print more debugging information into logs
    debug = CBool(False)

    #: (property) Is delayed change taking place at the moment?
    changing = Property(trait=Bool, transient=True, depends_on='_timed_action, _queued_job')

    @cached_property
    def _get_changing(self):
        if self._queued_job or self._timed_action:
            return True
        return False

    # Thread of currently running action
    _timed_action = Instance(TimerClass, transient=True)

    # Reference of status change job that is in the worker queue is saved here
    _queued_job = Instance(StatusWorkerTask, transient=True)

    # Time when status was last changed
    _last_changed = CFloat

    # The time when last change started
    _change_start = CFloat(transient=True)

    # Lock that is acquired when changing the status
    _status_lock = Instance(Lock, transient=True)

    logger = Instance(logging.Logger, transient=True)

    view = ["name", "status", "description", "safety_delay",
            "safety_mode", "change_delay", "change_mode"] + SystemObject.view

    simple_view = []

    # used by Web UI, for templates
    data_type = Str(transient=True)

    def __init__(self, *args, **kwargs):
        self._status_lock = Lock('statuslock')
        super(StatusObject, self).__init__(*args, **kwargs)

    def __setstate__(self, *args, **kwargs):
        self._status_lock = Lock('statuslock')
        super(StatusObject, self).__setstate__(*args, **kwargs)

    @property
    def is_program(self):
        """
            A property which can be used to check if StatusObject uses program features or not.
        """
        from automate.callables import Empty
        return not (isinstance(self.on_activate, Empty)
                    and isinstance(self.on_deactivate, Empty)
                    and isinstance(self.on_update, Empty))

    #: Status of the object.
    status = Property(depends_on="_status, _status_trigger", transient=True)

    #: To force trigger status change events even if status itself does not change
    _status_trigger = Event

    def get_status_display(self, **kwargs):
        """
            Define how status is displayed in UIs (add units etc.).
        """
        if 'value' in kwargs:
            value = kwargs['value']
        else:
            value = self.status

        return u"%s" % value

    def get_as_datadict(self):
        """
            Get data of this object as a data dictionary. Used by websocket service.
        """
        d = super(StatusObject, self).get_as_datadict()
        d.update(dict(status=self.status, data_type=self.data_type, editable=self.editable))
        return d

    @cached_property
    def _get_status(self):
        return self._status

    def _set_status(self, value):
        self.set_status(value)

    def setup_system(self, *args, **kwargs):
        super(StatusObject, self).setup_system(*args, **kwargs)
        self.data_type = self._status.__class__.__name__

    def set_status(self, new_status, origin=None, force=False):
        """
            For sensors, this is synonymous to::

                sensor.status = new_status

            For (non-slave) actuators, origin argument (i.e. is the program that is
            changing the status) need to be given,
        """
        raise NotImplementedError

    def update_status(self):
        """
           In sensors: implement particular value reading from device etc. here (this calls set_status(value)).
           In actuators: set value in particular device.
           Implement in subclasses.
        """

    def activate_program(self, program):
        """
            When program controlling this object activates, it calls this function.
        """

    def deactivate_program(self, program):
        """
            When program controlling this object deactivates, it calls this function.
        """

    def get_program_status(self, program):
        """
            Determine status of this object set by a particular program.
            Useful only for Actuators but defined here for interface compatibility.
        """
        return self.status

    def _set_real_status(self, status, prog):
        # _status_lock is acquired at StatusWorkerTask. Do not call this function from anywhere else!
        # This function is finally called by queue and this sets the value as desired.
        # Thus no need to worry that this function would be called in two threads
        # simultaneously (caller always WorkerThread).
        # prog is always None for non-actuators

        self.logger.debug('_set_real_status %s', status)
        if not self.silent:
            if isinstance(self, AbstractActuator):
                self.logger.info(u"%s %s (by %s) setting status to %s", self.__class__.__name__, self, prog, status)

        # if not self.slave and prog != getattr(self, 'program', None):
        if not getattr(self, 'slave', True) and not prog is getattr(self, 'program', None):
            self.logger.debug('program had changed, not changing')
            return

        self._change_start = 0.
        self._last_changed = time.time()

        try:
            if self._status == status:
                self._status_trigger = True
            else:
                self._status = status
        except TraitError as e:
            self.logger.warning('Wrong type of status %s was passed to %s. Error: %s', status, self, e)

    def _add_statuschange_to_queue(self, status, prog, with_statuslock=False):
        # This function is used for delayed actions
        if self.system.two_phase_queue:
            with self._status_lock if with_statuslock else nomutex:
                if status != self._status:
                    self._queued_job = StatusWorkerTask(func=self._set_real_status, args=(status, prog), object=self)
                    self.system.worker_thread.put(self._queued_job)
        else:
            self.system.worker_thread.put(DummyStatusWorkerTask(self._set_real_status, status, prog))

    def _are_delays_active(self, new_status):
        try:
            mode = "rising" if new_status > self._status else "falling"
        except TypeError:
            # if we can not determine whether value is rising or falling, we'll assume it's rising
            mode = 'rising'
        safety_active = False
        change_active = False
        if self.safety_delay > 0. and self.safety_mode in [mode, "both"]:
            safety_active = True
        if self.change_delay > 0. and self.change_mode in [mode, "both"]:
            change_active = True
        return safety_active, change_active

    def _do_change_status(self, status, force=False):
        """
        This function is called by
           - set_status
           - _update_program_stack if active program is being changed
             - thia may be launched by sensor status change.
             status lock is necessary because these happen from different
             threads.

        This does not directly change status, but adds change request
        to queue.
        """
        self.system.worker_thread.put(DummyStatusWorkerTask(self._request_status_change_in_queue, status, force=force))

    @property
    def next_scheduled_action(self):
        return getattr(self._timed_action, 'next_action', None)

    def _request_status_change_in_queue(self, status, force=False):
        def timer_func(func, *args):
            func(*args)
            self._timed_action = None

        with self._status_lock:
            timenow = time.time()
            logger = self.logger.info if self.debug else self.logger.debug
            logger("_do_change_status prg:%s status:%s", getattr(self, 'program', None), status)

            if self._timed_action:
                if self._timed_action.is_alive():
                    logger("Cancelling previous safety/change_delay action. Now changing to %s", status)
                    self._timed_action.cancel()
                self._timed_action = None

            self._queued_job = None

            safetydelay_active, changedelay_active = self._are_delays_active(status)
            run_now = not (safetydelay_active or changedelay_active)

            if not self._change_start:
                self._change_start = timenow

            if status == self._status and not force:
                self._change_start = 0
                logger('Status same %s, no need to change', status)
                return

            orig_changedelay = self.change_delay if changedelay_active else 0
            safetydelay = self.safety_delay if safetydelay_active else 0

            changedelay = orig_changedelay - (timenow - self._change_start)

            if changedelay < 0:
                changedelay = orig_changedelay

            if run_now or (changedelay <= 0. and (timenow - self._last_changed > safetydelay)):
                logger("Adding status change to queue, about to change status to %s", status)
                if self.system.two_phase_queue:
                    self._add_statuschange_to_queue(status, getattr(self, "program", None))
                else:
                    self._set_real_status(status, getattr(self, 'program', None))

            else:
                timesince = time.time() - self._last_changed
                delaytime = max(0, safetydelay - timesince, changedelay)
                time_after_delay = datetime.datetime.now() + datetime.timedelta(seconds=delaytime)
                logger("Scheduling safety/change_delay timer for %f sek. Now %s. Going to change to %s.",
                       delaytime, self._status, status)
                self._timed_action = threading.Timer(delaytime, timer_func,
                                                     args=(self._add_statuschange_to_queue, status, getattr(self, "program", None), True))
                self._timed_action.name = "Safety/change_delay for %s timed at %s (%f sek)" % (self.name, time_after_delay, delaytime)
                self._timed_action.next_action = time_after_delay
                self._timed_action.start()
                return False


class AbstractSensor(StatusObject):

    """ Base class for all sensors """

    #: Is sensor user-editable in UIs. This variable is meant for per-instance tuning for Sensors,
    #: whereas :attr:`.editable` is for per-class adjustment.
    user_editable = CBool(False)

    #: Status type is defined as _status, but always use :attr:`.status` property to set/get values.
    _status = Any

    #: Default value for status
    default = Any

    #: If non-zero, Sensor status will be reset to default after defined time (in seconds).
    reset_delay = CFloat

    _reset_timer = Any(transient=True)

    #: Do not log status changes
    silent = CBool(False)

    view = StatusObject.view + ['default', 'name', 'tags', 'reset_delay']
    simpleview = StatusObject.simple_view + ['_status']

    def get_as_datadict(self):
        d = super(AbstractSensor, self).get_as_datadict()
        d.update(dict(user_editable=self.user_editable))
        return d

    def _setup_reset_delay(self):
        if self.reset_delay:
            if self._reset_timer and self._reset_timer.is_alive():
                self._reset_timer.cancel()
            self._reset_timer = threading.Timer(self.reset_delay, lambda: self.set_status(self.default))
            self._reset_timer.start()

    def set_status(self, status, origin=None, force=False):
        """
            Compatibility to actuator class.
            Also :class:`~automate.callables.builtin_callables.SetStatus`
            callable can be used for sensors too, if so desired.
        """

        if status != self.default:
            self._setup_reset_delay()
        return self._do_change_status(status, force)

    def update_status(self):
        """A method to read and update actual status. Implement it in subclasses, if necessary"""

    def _status_changed(self):
        if not self.silent:
            self.logger.info("%s status changed to %s", self, repr(self.status))

    def __str__(self):
        return self.name

    def setup_system(self, system, *args, **kwargs):
        name, traits = self._passed_arguments
        self.logger = logging.getLogger('automate.sensor')
        default = traits.get('default', None)
        super(AbstractSensor, self).setup_system(system, *args, **kwargs)
        if not default is None and not kwargs.get('loadstate', None):
            self.set_status(default)


class AbstractActuator(StatusObject):

    """ Base class for all actuators."""

    #: Default value for status. For actuators, this is set by automatically created
    #: :class:`~automate.program.DefaultProgram` dp_actuatorname
    default = Any(False, transient=True)

    #: If ``True``, actual status can be set by any program anytime without restrictions.
    slave = CBool

    #: A property giving current program governing the status of this actuator (program that has the highest priority)
    program = Property(trait=Instance(ProgrammableSystemObject), transient=True, depends_on="program_stack[]")

    #: This dictionary can be used to override program priorities.
    #:
    #: .. note::  Keys here must be program names, (not :class:`~automate.program.Program` instances).
    priorities = Dict(key_trait=Str, value_trait=CFloat)

    @cached_property
    def _get_program(self):
        try:
            return self.program_stack[-1]
        except IndexError:
            return

    #: Reference to actuators :class:`~automate.program.DefaultProgram`
    default_program = Instance(DefaultProgram)

    # Locks for thread-safety
    _actuator_status_lock = Instance(Lock, transient=True)

    _program_lock = Instance(Lock, transient=True)

    view = StatusObject.view + ["status", "slave", "tags"]
    simple_view = StatusObject.simple_view + ['status']

    def __init__(self, *args, **kwargs):
        self._actuator_status_lock = Lock("statuslock")
        self._program_lock = Lock("programlock")
        super(AbstractActuator, self).__init__(*args, **kwargs)

    def __setstate__(self, *args, **kwargs):
        self._actuator_status_lock = Lock("statuslock")
        self._program_lock = Lock("programlock")
        return super(AbstractActuator, self).__setstate__(*args, **kwargs)

    def set_status(self, status, origin=None, force=False):
        """ For programs, to set current status of the actuator. Each
            active program has its status in :attr:`.program_stack`
            dictionary and the highest priority is realized in the actuator """

        if not self.slave and origin not in self.program_stack:
            raise ValueError('Status cannot be changed directly')

        with self._actuator_status_lock:
            self.logger.debug("set_status got through, program: %s", origin)
            if self.debug:
                self.logger.info("Set_status %s %s %s", self.name, origin, status)

            if self.slave:
                return self._do_change_status(status, force)

            self.logger.debug("Sets status %s for %s", status, origin.name)

            with self._program_lock:
                self.program_status[origin] = status

                if self.program == origin:
                    return self._do_change_status(status, force)

    def activate_program(self, program):
        """
            Called by program which desires to manipulate this actuator, when it is activated.
        """
        self.logger.debug("activate_program %s", program)
        if program in self.program_stack:
            return

        with self._program_lock:
            self.logger.debug("activate_program got through %s", program)
            self.program_stack.append(program)
            self._update_program_stack()

    def deactivate_program(self, program):
        """
            Called by program, when it is deactivated.
        """
        self.logger.debug("deactivate_program %s", program)

        with self._program_lock:
            self.logger.debug("deactivate_program got through %s", program)
            if program not in self.program_stack:
                import ipdb
                ipdb.set_trace()
            self.program_stack.remove(program)
            if program in self.program_status:
                del self.program_status[program]
            self._update_program_stack()

    def update_program_stack(self):
        """
            Update :attr:`.program_stack`. Used by programs
            :attr:`~automate.program.ProgrammableSystemObject._priority_changed` attribute to reset ordering.
        """
        with self._program_lock:
            self._update_program_stack()

    def get_program_status(self, prog):
        """
            Give status defined by given program ``prog``
        """
        try:
            return self.program_status[prog]
        except KeyError:
            return None

    def _status_changed(self):
        """
            Override this method to whatever action you want to actualize when
            status is changed. This one is called in the initialization procedure as well.

            Usually this method is enough. Sometimes you might not want to do anything
            in the initialization, and then you can override __status_changed method instead.
        """

    # Private attributes

    #: Use property 'status' instead in normal operations (status is set via status worker thread).
    _status = Any(transient=True)

    #: Program stack of current programs, sorted automatically by program priority.
    program_stack = List(trait=Instance(ProgrammableSystemObject), transient=True)

    #: Dictionary containing statuses set by each active program
    program_status = Dict(key_trait=Instance(ProgrammableSystemObject), transient=True)

    def setup_system(self, system, *args, **kwargs):
        from automate.callables import SetStatus, Value, Empty
        super(AbstractActuator, self).setup_system(system, *args, **kwargs)
        actevent = SetStatus(self, Value(self.default))
        if self.slave:
            actevent = Empty()
            self.status = self.default

        from .program import DefaultProgram

        if not self.default_program:
            self.default_program = DefaultProgram(
                targets=[self],
                active_condition=Value(False),
                on_activate=actevent,
                priority=0,
                system=self.system,
                name='dp_%s' % self.name,
            )

            self.default_program.active_condition = Value(True)

        self._program_lock.name = "programlock " + self.name
        self._actuator_status_lock.name = "statuslock " + self.name

    def __str__(self):
        return self.name

    def _update_program_stack(self):
        # Use of this function must be protected from outside by self.program_lock!
        if self.priorities:
            self.program_stack.sort(key=lambda x: self.priorities.get(x.name, x.priority))
        else:
            self.program_stack.sort(key=operator.attrgetter('priority'))
        if self.program in self.program_status:
            self._do_change_status(self.program_status[self.program])



# -*- coding: utf-8 -*-
# (c) 2015 Tuomas Airaksinen
#
# This file is part of automate-webui.
#
# automate-webui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# automate-webui is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with automate-webui.  If not, see <http://www.gnu.org/licenses/>.
#
# ------------------------------------------------------------------
#
# If you like Automate, please take a look at this page:
# http://evankelista.net/automate/

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals
import threading
from django.contrib import messages

from django.core.urlresolvers import reverse
from django.http import HttpResponseRedirect, Http404, HttpResponse
from django.shortcuts import render, redirect
from django.template import Template, RequestContext
from django.utils.http import urlencode
from django.utils.safestring import mark_safe
from .forms import LoginForm, CmdForm, FORMTYPES, QUICK_EDITS, TextForm
from functools import wraps
from automate.statusobject import AbstractActuator
from automate.statusobject import AbstractSensor


def set_globals(_service, _system):
    global system, service
    system = _system
    service = _service


def get_groups(only_user_editable=False, only_user_defined=False, only_groups=False):
    groups = {}
    if only_user_editable:
        objs = (i for i in service.system.objects_sorted if getattr(i, 'user_editable', False))
    elif only_user_defined:
        objs = (i for i in service.system.objects_sorted if service.user_tags & i.tags)
    else:
        objs = service.system.objects_sorted
    if not service.show_hidden:
        objs = (i for i in objs if not i.name.startswith('_'))
    for obj in objs:
        for tag in obj.tags:
            if only_groups:
                if not tag.startswith('group:'):
                    continue
            if tag in groups:
                groups[tag].append(obj)
            else:
                groups[tag] = [obj]
    return groups


def common_context(request):
    global system, service
    _views = [
        ('By types', 'system'),
        ('By groups', 'tag_view_only_groups'),
        ('By tags', 'tags_view'),
        ('User editable', 'user_editable_view'),
        ('User defined', 'user_defined_view'),
        ('System and threads', 'threads'),
        ('Console', 'console'),
        ('UML', 'plantuml'),
    ]
    current_view = request.resolver_match.view_name
    views = [(title, reverse(name), name == current_view) for title, name in _views]

    groups = get_groups()

    tag_views = [(name, reverse('single_tag', args=(name,)), False)
                 for name in sorted(groups.keys())]

    from automate import __version__
    return {'views': views, 'tag_views': tag_views, 'system': system, 'service': service,
            'automate_version': __version__}


def require_login(func):
    @wraps(func)
    def wrapped(request, *args, **kwargs):
        if not request.session.get('logged_in', False):
            service.system.logger.debug('Require login')
            return HttpResponseRedirect(
                reverse('login') + '?%s' % urlencode({'url': request.META['PATH_INFO']}))
        service.system.flush()
        return func(request, *args, **kwargs)

    return wrapped


def login(request):
    url = request.GET.get('url', reverse('main'))
    if not service.http_auth or request.session.get('logged_in', False):
        request.session['logged_in'] = True
        return HttpResponseRedirect(url)
    if request.method == 'POST':
        form = LoginForm(request.POST, request=request)
        form.auth = service.http_auth
        if form.is_valid():
            request.session['logged_in'] = True
            return HttpResponseRedirect(url)
    else:
        form = LoginForm(request=request)
    form.helper.form_class = 'form-signin'
    form.helper.form_action += '?url=%s' % url
    return render(request, 'login.html', {'form': form, 'source': 'login'})


@require_login
def main(request):
    return HttpResponseRedirect(reverse(service.default_view))


def logout(request):
    request.session['logged_in'] = False
    return HttpResponseRedirect('login')


@require_login
def custom(request, name):
    context = RequestContext(request, {'source': 'main'})
    try:
        return HttpResponse(content=Template(service.custom_pages[name]).render(context))
    except KeyError:
        raise Http404


@require_login
def puml_svg(request):
    svg = service.system.request_service('PlantUMLService').write_svg()
    return HttpResponse(content=svg, content_type='image/svg+xml')


@require_login
def puml_raw(request):
    puml = service.system.request_service('PlantUMLService').write_puml()
    return HttpResponse(content=puml, content_type='text/plain')


@require_login
def plantuml(request):
    puml_service = service.system.request_service('PlantUMLService')
    return render(request, 'views/plantuml.html', {'source': 'system',
                                                   'puml_service': puml_service,
                                                   })


@require_login
def system_view(request):
    return render(request, 'views/system.html', {'source': 'system'})


@require_login
def threads(request):
    threads = [(t.name, t) for t in threading.enumerate()]
    threads.sort()
    return render(request, 'views/threads.html', {'threads': threads})


@require_login
def single_tag(request, tag):
    objs = sorted([obj for obj in service.system.objects_sorted if tag in obj.tags])
    if not service.show_hidden:
        objs = (i for i in objs if not i.name.startswith('_'))

    if not objs:
        raise Http404
    return render(request, 'views/single_tag_view.html',
                  {'source': 'tags_view', 'objs': objs, 'tag': tag})



@require_login
def tags_view(request, template='', only_user_editable=False, only_user_defined=False,
              only_groups=False):
    groups = get_groups(only_user_editable=only_user_editable,
                        only_user_defined=only_user_defined, only_groups=only_groups)
    gitems = sorted(groups.items())
    l = len(gitems)
    g1 = gitems[:int(l / 3) + 1]
    g2 = gitems[int(l / 3) + 1:2 * int(l / 3) + 1]
    g3 = gitems[2 * int(l / 3) + 1:]
    return render(request, 'views/tag_view.html' if not template else template,
                  {'source': 'user_editable_view' if only_user_editable else 'tags_view',
                   'groups': [g1, g2, g3]})


@require_login
def user_editable_view(request):
    return tags_view(request, template='views/user_editable_view.html', only_user_editable=True)


@require_login
def user_defined_view(request):
    return tags_view(request, template='views/user_defined_view.html', only_groups=True,
                     only_user_defined=True)


@require_login
def tag_view_only_groups(request):
    return tags_view(request, template='views/only_groups.html', only_groups=True)


@require_login
def info_panel(request, name):
    source = request.GET.get('source', 'main')
    if request.is_ajax():
        obj = service.system.namespace[name]
        view_items = obj.view[:] + ['class_name', 'data_type', 'next_scheduled_action']
        if 'change_delay' in view_items and not obj.change_delay:
            view_items.remove('change_delay')
            view_items.remove('change_mode')
        if 'safety_delay' in view_items and not obj.safety_delay:
            view_items.remove('safety_delay')
            view_items.remove('safety_mode')
        if 'reset_delay' in view_items and not obj.reset_delay:
            view_items.remove('reset_delay')
        info_items = [(i.capitalize().replace('_', ' '),
                       getattr(obj, i)) for i in view_items
                      if (not i.endswith('_str')
                          and i not in ['tags', 'name', 'priority', 'status']
                          and (
                          getattr(obj, i, None) or type(getattr(obj, i, None)) in (int, float)))]

        callables = ((i.capitalize().replace('_', ' '), i) for i in obj.callables)

        textform = TextForm({'status': obj.status, 'name': obj.name},
                            source=source) if obj.data_type in ['str', 'unicode'] else None

        return render(request, 'info_panel.html',
                      {'i': obj, 'source': source, 'info_items': info_items,
                       'callables': callables, 'textform': textform})
    else:
        raise Http404


@require_login
def toggle_sensor(request, sensorname):
    """
    This is used only if websocket fails
    """
    if service.read_only:
        service.logger.warning("Could not perform operation: read only mode enabled")
        raise Http404
    source = request.GET.get('source', 'main')
    sensor = service.system.namespace[sensorname]
    sensor.status = not sensor.status
    service.system.flush()
    return HttpResponseRedirect(reverse(source))


@require_login
def toggle_value(request, name):
    """
    For manual shortcut links to perform toggle actions
    """
    obj = service.system.namespace.get(name, None)
    if not obj or service.read_only:
        raise Http404
    new_status = obj.status = not obj.status
    if service.redirect_from_setters:
        return HttpResponseRedirect(reverse('set_ready', args=(name, new_status)))
    else:
        return set_ready(request, name, new_status)


@require_login
def set_value(request, name, value):
    """
    For manual shortcut links to perform set value actions
    """
    obj = service.system.namespace.get(name, None)
    if not obj or service.read_only:
        raise Http404
    obj.status = value
    if service.redirect_from_setters:
        return HttpResponseRedirect(reverse('set_ready', args=(name, value)))
    else:
        return set_ready(request, name, value)


@require_login
def set_ready(request, name, value):
    return render(request, 'views/toggle.html', {'name': name, 'status': value})


@require_login
def edit(request, name=None, type=None, cont=False):
    if service.read_only:
        service.logger.warning("Could not perform operation: read only mode enabled")
        raise Http404

    source = request.GET.get('source', 'main')
    if name:
        obj = service.system.namespace.get(name, None)
        if not obj:
            raise Http404
        if isinstance(obj, AbstractActuator):
            type = 'actuator'
        elif isinstance(obj, AbstractSensor):
            type = 'sensor'
        else:
            type = 'program'

    if not type:
        raise Http404

    FormType = FORMTYPES[type]

    if request.method == 'POST':
        if 'cancel' in request.POST:
            return HttpResponseRedirect(reverse(source))

        form = FormType(request.POST, request=request)
        form.setup_system(service.system)
        form.load(name)
        if form.is_valid():
            if form.has_changed():
                form.save(name)
            if name:
                return HttpResponseRedirect(reverse(source))
            else:
                return HttpResponseRedirect(
                    reverse('continue_edit', args=(type, form.instance.name,))
                    + '?source=%s' % source)
    else:
        form = FormType(request=request)
        form.setup_system(service.system)
        form.load(name)
    return render(request, 'edit_object.html', {'type': type, 'editform': form,
                                                "continue_edit": cont, 'source': source})


@require_login
def continue_edit(request, type, name):
    if service.read_only:
        service.logger.warning("Could not perform operation: read only mode enabled")
        raise Http404
    return edit(request, type=type, name=name, cont=True)


@require_login
def new(request, type):
    if service.read_only:
        service.logger.warning("Could not perform operation: read only mode enabled")
        raise Http404
    return edit(request, type=type, name=None)


@require_login
def console(request):
    textarea = False
    if request.method == 'POST':
        cmdform = CmdForm(request.POST, textarea=True)
        if cmdform.is_valid():
            if not service.read_only:
                service.system.cmd_exec(cmdform.cleaned_data['cmd'])
            else:
                service.logger.warning("Could not perform operation: read only mode enabled")
            return HttpResponseRedirect('')
    else:
        textarea = request.GET.get('textarea', False)
        cmdform = CmdForm(textarea=textarea)

    log = mark_safe(service.system.request_service('LogStoreService').lastlog(lines=100))
    return render(request, 'views/console.html', dict(log=log, cmdform=cmdform, textarea=textarea))


@require_login
def set_status(request):
    if service.read_only:
        service.logger.warning("Could not perform operation: read only mode enabled")
        raise Http404
    if request.method == 'POST':
        name = request.POST.get('name', None)
        status = request.POST.get('status', None)
        obj = service.system.namespace.get(name, None)
        if obj:
            service.logger.debug('POST: %s', request.POST)
            form = QUICK_EDITS[obj.data_type](request.POST, sensor=obj)
            if form.is_valid():
                obj.status = form.cleaned_data['status']
                service.system.flush()
            else:
                messages.error(request, 'Error setting status %s to %s' % (name, status))
        else:
            messages.error(request, 'Error setting status of %s (no such object)' % name)

        return HttpResponseRedirect(reverse(request.GET.get('source', 'main')))
    else:
        raise Http404


@require_login
def reload_service(request, id_):
    id_ = int(id_)
    found = False
    for ser in system.services:
        if id(ser) == id_:
            ser.reload()
            messages.info(request, 'Reloaded service %s (%s)' % (ser.__class__.__name__, ser.id))
            found = True
            break
    if not found:
        messages.error(request, 'Service was not found')

    return redirect(request.META.get('HTTP_REFERER', reverse('main')))


@require_login
def cancel_thread(request, id_):
    id_ = int(id_)
    found = False
    for t in threading.enumerate():
        if id(t) == id_:
            found = True
            try:
                t.cancel()
                messages.info(request, 'Canceled thread %s (%s)' % (t, id_))
            except AttributeError:
                messages.error(request, 'Could not cancel thread %s (%s)' % (t, id_))
            break
    if not found:
        messages.error(request, 'Thread %s was not found' % id_)

    return redirect(request.META.get('HTTP_REFERER', reverse('main')))

def notfound(request):
    raise Http404



# -*- coding: utf-8 -*-
# (c) 2015 Tuomas Airaksinen
#
# This file is part of Automate.
#
# Automate is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Automate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Automate.  If not, see <http://www.gnu.org/licenses/>.
#
# ------------------------------------------------------------------
#
# If you like Automate, please take a look at this page:
# http://evankelista.net/automate/

from __future__ import unicode_literals
from past.builtins import basestring
import re
from traits.api import cached_property, on_trait_change, CList, Dict, Instance, Set, Event, Property
from automate.common import CompareMixin, Lock, deep_iterate, Object, is_iterable, AbstractStatusObject, DictObject, SystemNotReady
from automate.systemobject import SystemObject

import logging


class AbstractCallable(SystemObject, CompareMixin):

    """

    A base class for subclassing Callables that are used in Program conditions and action attributes.

    Callables are configured by giving them arguments and keyword arguments.They must always define :meth:`.call`
    method which defines their functionality.
    """

    #: Arguments given for callable are stored here
    _args = CList

    #: Keyword arguments given for callable are stored here
    _kwargs = Dict

    #: Lock must be used when accessing :attr:`.state`
    _lock = Instance(Lock, transient=True)

    #: Property that gives set of all *triggers* of this callable and it's children callables.
    #: Triggers are all those StatusObjects that alter the status (return value of :meth:`.call`) of
    #: Callable.
    triggers = Property(trait=Set(trait=AbstractStatusObject),
                        depends_on='_args, _args_items, _args.triggers, _kwargs, _kwargs_items, _kwargs.triggers')

    #: Property that gives set of all *targets* of this callable and it's children callables. Targets are
    #: all those StatusObjects of which status the callable might alter in :meth:`.call`.
    targets = Property(trait=Set(trait=AbstractStatusObject),
                       depends_on='_args, _args_items, _args.targets, _kwargs, _kwargs_items, _kwargs.targets')

    # Status property not yet used anywhere. In the future, Program.active could be changed to Property
    # which uses this status.
    # Remark: isn't it used in web?

    #: Read-only status property of the callable. Usefull only when callable is used as a condition.
    #: This automatically depends on all the StatusObjects below the Callable tree.
    status = Property(depends_on='_args.status, _kwargs.status')

    #: State dictionary that is used by :meth:`.call` and :meth:`.cancel` if some state variables are needed to be saved
    #: Remember to clean data in subclasses when it is no longer needed.
    state = Dict(transient=True)

    def get_state(self, caller):
        """
            Get per-program state.
        """

        if caller in self.state:
            return self.state[caller]
        else:
            rv = self.state[caller] = DictObject()
            return rv

    def del_state(self, caller):
        """
            Delete per-program state.
        """
        if caller in self.state:
            del self.state[caller]

    @cached_property
    def _get_status(self):
        if not self.system:
            # Raising exception prevents invalid value from being cached when system is in pre-mature state
            raise SystemNotReady('System not ready yet -- this is normal when loading dump.')
        return self.call(None)

    #: Event that can be used to execute code right after callable setup. See :class:`.OfType`.
    #: Something that needs to be done manually this way, because Traits does not allow
    #: defining the order of subscribed function calls.
    on_setup_callable = Event

    def setup_callables(self):
        super(AbstractCallable, self).setup_callables()
        self.setup_callable_system(self.system, init=True)

    @on_trait_change('_args, _args_items', post_init=True)
    def objects_changed(self, name, old, new):
        if not self.system:
            return
        for o in new.added:
            if isinstance(o, AbstractCallable):
                o.setup_callable_system(self.system)

    @on_trait_change('_kwargs, _kwargs_items', post_init=True)
    def kwargs_changed(self, name, old, new):
        if not self.system:
            return
        for o in list(new.added.values()):
            if isinstance(o, AbstractCallable):
                o.setup_callable_system(self.system)

    @cached_property
    def _get_triggers(self):
        if not self.system:
            # Raising exception prevents invalid value from being cached when system is in pre-mature state
            raise SystemNotReady('System not ready yet -- this is normal when loading dump.')
        return self.collect('triggers')

    @cached_property
    def _get_targets(self):
        if not self.system:
            # Raising exception prevents invalid value from being cached when system is in pre-mature state
            raise SystemNotReady('System not ready yet -- this is normal when loading dump.')
        return self.collect('targets')

    def __getitem__(self, item):
        return self._args[item]

    def __setitem__(self, item, value):
        self._args[item] = value

    def __init__(self, *args, **kwargs):
        self._lock = Lock("Lock for callable " + self.__class__.__name__)
        self._kwargs = kwargs
        super(AbstractCallable, self).__init__()
        super(SystemObject, self).__init__()
        if not self.traits_inited():
            self.logger.error('Traits not inited!!!')
        if args:
            self._args = args

    def __setstate__(self, state, trait_change_notify=True):
        self._lock = Lock("Lock for callable ")
        self._passed_arguments = None, state.copy()
        self.logger = logging.getLogger('automate.%s' % self.__class__.__name__)
        state.pop('name', '')
        super(SystemObject, self).__setstate__(state, trait_change_notify)

    def call_eval(self, value, caller, return_value=True, **kwargs):
        """
            Value might be either name registered in System namespace, or object, either
            StatusObject or Callable. If Callable, evaluate :meth:`.call` method. If StatusObject,
            return status.
        """
        value = self.name_to_system_object(value)
        if return_value and isinstance(value, AbstractStatusObject):
            return value.status
        if hasattr(value, 'call'):
            return self.call_eval(value.call(caller, **kwargs), caller, return_value, **kwargs)
        else:
            return value

    def _fix_list(self, lst):
        if isinstance(lst, dict):
            lst2 = list(lst.items())
        elif isinstance(lst, list):
            lst2 = enumerate(lst)
        else:
            raise RuntimeError('Error in _fix_list, type %s', type(lst))
        for idx, obj in lst2:
            if isinstance(obj, basestring):
                lst[idx] = self.name_to_system_object(obj)
            if isinstance(obj, list):
                self._fix_list(obj)

    def setup_callable_system(self, system, init=False):
        """
            This function basically sets up :attr:`.system`, if it is not yet set up. After that,
            other Callable initialization actions are performed.

            :param init: value ``True`` is given when running this at the initialization phase. Then system
                         attribute is set already, but callable needs to be initialized otherwise.

        """
        if not self.system or init:
            self.system = system
            self._fix_list(self._args)
            self._fix_list(self._kwargs)
            for i in self.children:
                if isinstance(i, AbstractCallable):
                    i.setup_callable_system(system, init=init)
                elif isinstance(i, SystemObject):
                    i.system = system
            self.on_setup_callable = 1

    def call(self, *args, **kwargs):
        """
            The basic functionality of the Callable is implemented in this function.
            Needs to be defined in derived subclasses.

            If callable is used as a Program condition, this must return the value of the condition
            (see for example conditions :class:`.And`, :class:`.Sum` etc.), otherwise return value is optional.
        """
        raise NotImplementedError

    @property
    def objects(self):
        """
            Shortcut to :attr:`._args`.
        """
        return self._args

    @property
    def obj(self):
        """
            Shortcut property to the first stored object.
        """
        try:
            return self._args[0]
        except IndexError:
            return None

    @property
    def value(self):
        """
            Shortcut property to the second stored object.
        """
        try:
            return self._args[1]
        except IndexError:
            return None

    def name_to_system_object(self, value):
        """
        Return object for given name registered in System namespace.
        """
        if not self.system:
            raise SystemNotReady

        if isinstance(value, (basestring, Object)):
            rv = self.system.name_to_system_object(value)
            return rv if rv else value
        else:
            return value

    def collect(self, target):
        """Recursively collect all potential triggers/targets in this node and its children.
        Define targets and triggers of this particular callable in :meth:`_give_triggers`
        and :meth:`_give_targets`.

        :param str target: valid values: ``'targets'`` and ``'triggers'``
        """
        statusobjects = set()
        callables = set()
        objs_from_this_obj = getattr(self, '_give_%s' % target)()

        if not is_iterable(objs_from_this_obj):
            objs_from_this_obj = [objs_from_this_obj]

        if is_iterable(objs_from_this_obj):
            for i in (self.name_to_system_object(j) for j in objs_from_this_obj):
                if isinstance(i, AbstractStatusObject):
                    statusobjects.add(i)
                elif isinstance(i, AbstractCallable):
                    callables.add(i)

        for i in (self.name_to_system_object(j) for j in deep_iterate(callables)):
            if isinstance(i, AbstractCallable):
                statusobjects.update(getattr(i, target))

        return statusobjects

    def collect_triggers(self):
        return self.collect('triggers')

    def collect_targets(self):
        return self.collect('targets')

    @property
    def children(self):
        """
            A property giving a generator that goes through all the children of this Callable (not recursive)
        """
        return deep_iterate(self._args + list(self._kwargs.values()))

    def _give_triggers(self):
        """Give all triggers of this object (non-recursive)"""
        return self.children

    def _give_targets(self):
        """Give all targets of this object (non-recursive)"""
        return self.children

    def cancel(self, caller):
        """
            Recursively cancel all threaded background processes of this Callable.
            This is called automatically for actions if program deactivates.
        """
        for o in {i for i in self.children if isinstance(i, AbstractCallable)}:
            o.cancel(caller)

    def __repr__(self):
        return self.give_str()

    def __str__(self):
        return self.give_str()

    def _give_str(self, args, kwargs):
        if self in self.system.namespace.reverse:
            return repr(self.name)
        kwstr = u', '.join(k + u'=' + repr(v) for k, v in list(kwargs.items()))
        if kwstr and args:
            kwstr = ', ' + kwstr
        return str(self.__class__.__name__) + u"(" + u", ".join([repr(i) for i in args]) + kwstr + u")"

    def give_str(self):
        """
            Give string representation of the callable.
        """
        args = self._args[:]
        kwargs = self._kwargs
        return self._give_str(args, kwargs)

    @staticmethod
    def strip_color_tags(_str):
        return re.sub('__\w*__', "", _str)

    def _give_str_indented(self, args, kwargs, tags):
        if self in self.system.namespace.reverse:
            rv = repr(self.name)

        def indent(o_str):
            n_strs = []
            for o in o_str.split('\n'):
                n_strs.append(u'  ' + o)
            return '\n'.join(n_strs)

        def indented_str(obj, no_repr=False, no_color=False):
            if hasattr(obj, 'give_str_indented') and not obj in self.system.namespace.reverse:
                rv = obj.give_str_indented(tags)
            else:
                rv = str(obj if no_repr else repr(obj))
                if not no_color:
                    rv = ('__ACT__' if getattr(obj, 'status', obj) else '__INACT__') + rv
            return indent(rv)

        def in_one_line(obj):
            rv = repr(obj)
            if not isinstance(obj, AbstractCallable):
                rv = ('__ACT__' if getattr(obj, 'status', obj) else '__INACT__') + rv
            return rv

        kwstrs = [k + u'=' + in_one_line(v) for k, v in list(kwargs.items())]

        argstr = u"(\n" + indent(u", \n".join([indented_str(i) for i in args] +
                                              [indented_str(i, no_repr=True, no_color=True) for i in kwstrs]) + u"\n)")
        if len(self.strip_color_tags(argstr)) < 35:
            argstr = u"(" + u", ".join([in_one_line(i) for i in args] + kwstrs) + u")"

        rv = str(self.__class__.__name__) + argstr
        if tags:
            rv = ('__ACT__' if self.status else '__INACT__') + rv
        return rv

    def give_str_indented(self, tags=False):
        """
            Give indented string representation of the callable.
            This is used in :ref:`automate-webui`.
        """
        args = self._args[:]
        kwargs = self._kwargs
        rv = self._give_str_indented(args, kwargs, tags)
        if not tags:
            rv = self.strip_color_tags(rv)
        return rv

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return (self._args, self._kwargs) == (other._args, other._kwargs)
        return False

    def __hash__(self):
        return id(self)


#!/usr/bin/env python

from setuptools import setup, find_packages


# Extra requirements
rpc_requirements = [
    "tornado~=4.5.0",
]

web_requirements = [
    "Django~=1.11.0",
    "django-crispy-forms~=1.6.0",
    # "whitenoise",
] + rpc_requirements

gpio_requirements = ['RPi.GPIO']
rpio_requirements = ['RPIO']
arduino_requirements = ['pyfirmata']

all_extras_requirements = web_requirements + gpio_requirements + arduino_requirements

setupopts = dict(
    name="automate",
    version='0.10.12',
    packages=find_packages('src'),
    include_package_data=True,
    package_dir={'': 'src'},
    zip_safe=False,
    install_requires=[
        "ansiconv",
        "colorlog",
        "croniter~=0.3.0",
        "future",
        "ipython<6.0",
        "pyinotify",
        "traits~=4.6.0",
        "mock",
        "requests",
        ],
    extras_require={
        'web': web_requirements,
        'rpc': rpc_requirements,
        'raspberrypi': gpio_requirements,
        'rpio': rpio_requirements,
        'arduino': arduino_requirements,
        'all': all_extras_requirements,
    },

    download_url='https://pypi.python.org/pypi/automate',
    platforms = ['any'],
    author="Tuomas Airaksinen",
    author_email="tuomas.airaksinen@gmail.com",
    description="General purpose Python automatization library with nifty real-time web UI",
    long_description=open('README.rst').read(),
    license="GPL",
    keywords="automation, GPIO, Raspberry Pi, RPIO, traits",
    url="http://github.com/tuomas2/automate",

    classifiers= [
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Environment :: Web Environment",
        "Intended Audience :: Education",
        "Intended Audience :: End Users/Desktop",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Topic :: Scientific/Engineering",
        "Topic :: Software Development",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
    ],
    entry_points={
        'automate.extension': [
            'arduino = automate.extensions.arduino:extension_classes',
            'rpc = automate.extensions.rpc:extension_classes',
            'rpio = automate.extensions.rpio:extension_classes',
            'webui = automate.extensions.webui:extension_classes',
        ]
    },
)

if __name__ == "__main__":
    setup(**setupopts)



# -*- coding: utf-8 -*-
# (c) 2015 Tuomas Airaksinen
#
# This file is part of automate-rpio.
#
# automate-rpio is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# automate-rpio is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with automate-rpio.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals

from threading import Thread

from traits.api import CBool, Any
from automate.service import AbstractSystemService
import signal

class RpioService(AbstractSystemService):

    """
        Service that provides interface to Raspberry Pi GPIO via
        `RPIO library <http://pythonhosted.org/RPIO/>`_.
    """

    #: Perform GPIO cleanup when exiting (default: False).
    gpio_cleanup = CBool(False)

    #: Use RPIO instead of RPI.GPIO
    rpio = CBool(False)

    _gpio_thread = Any

    _hw = Any

    def setup(self):
        self.logger.info("Initializing RpioService (Raspberry Pi GPIO support)")
        try:
            if self.rpio:
                import RPIO
                import RPIO.PWM
            else:
                import RPi.GPIO as RPIO
            self.is_mocked = False
        except (ImportError, SystemError, RuntimeError):
            self.logger.warning('RPIO module could not be imported. Enabling mocked RPIO')
            self.logger.warning("To use Raspberry Pi GPIO ports (sensors / actuators) please install module RPIO")
            import mock
            RPIO = mock.MagicMock()
            self.is_mocked = True

        self._hw = RPIO

        self._hw.setmode(RPIO.BCM)
        if self.rpio:
            self._gpio_thread = t = Thread(target=RPIO.wait_for_interrupts, name='RpioService thread')
            t.daemon = True
            t.start()

        self.logger.info("RPIO initialized")

    def cleanup(self):

        if self.rpio:
            self._hw.stop_waiting_for_interrupts()
            self._hw.cleanup_interrupts()
            self._gpio_thread.join()

        if self.gpio_cleanup:
            self._hw.cleanup()

    def enable_input_port(self, port, callback, pull_up_down):
        pud = {"down": self._hw.PUD_DOWN, "up": self._hw.PUD_UP, "none": self._hw.PUD_OFF}
        if self.rpio:
            self._hw.setup(port, self._hw.IN)
            self._hw.add_interrupt_callback(port, callback, edge="both", pull_up_down=pud[pull_up_down])
        else:
            self._hw.setup(port, self._hw.IN, pull_up_down=pud[pull_up_down])
            self._hw.add_event_detect(port, self._hw.BOTH, lambda _port: callback(_port, self._hw.input(_port)))

    def get_input_status(self, port):
        return self._hw.input(port)

    def disable_input_port(self, port):
        if self.rpio:
            self._hw.del_interrupt_callback(port)
        else:
            self._hw.remove_event_detect(port)
            self._hw.cleanup(port)

    def enable_output_port(self, port):
        self._hw.setup(port, self._hw.OUT)

    def disable_output_port(self, port):
        self._hw.setup(port, self._hw.IN)

    def get_pwm_module(self):
        if not self.rpio:
            self.logger.error('PWM supported only when rpio attribute in RpioService is enabled')
        try:
            self._hw.PWM.setup()
            signal.signal(signal.SIGCHLD, signal.SIG_IGN)
            self.logger.warning('SIGCHLD is now ignored totally due to RPIO.PWM bug. Might cause side effects!')
        except RuntimeError as e:
            if 'has already been called before' in e.message:
                pass
            else:
                raise

        return self._hw.PWM

    def set_output_port_status(self, port, status):
        self._hw.output(port, status)



# -*- coding: utf-8 -*-
# (c) 2015 Tuomas Airaksinen
#
# This file is part of automate-arduino.
#
# automate-arduino is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# automate-arduino is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with automate-arduino.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals
from traits.api import CInt, Instance, CFloat, CBool
from automate.service import AbstractSystemService
from automate.statusobject import AbstractSensor


class AbstractArduinoSensor(AbstractSensor):

    """
        Abstract base class for Arduino sensors
    """

    user_editable = CBool(False)

    #: Arduino device number (specify, if more than 1 devices configured in ArduinoService)
    dev = CInt(0)

    #: Arduino pin number
    pin = CInt

    _arduino = Instance(AbstractSystemService, transient=True)

    def setup(self, *args, **kwargs):
        super(AbstractArduinoSensor, self).setup(*args, **kwargs)
        self._arduino = self.system.request_service('ArduinoService')


class ArduinoAnalogSensor(AbstractArduinoSensor):

    """
        Float-valued sensor object for analog Arduino input pins
    """
    _status = CFloat

    def setup(self, *args, **kwargs):
        super(ArduinoAnalogSensor, self).setup(*args, **kwargs)
        self._arduino.subscribe_analog(self.dev, self.pin, self)

    def cleanup(self):
        self._arduino.unsubscribe_analog(self.dev, self.pin)


class ArduinoDigitalSensor(AbstractArduinoSensor):

    """
        Boolean-valued sensor object for digital Arduino input pins
    """

    _status = CBool

    def setup(self, *args, **kwargs):
        super(ArduinoDigitalSensor, self).setup(*args, **kwargs)
        self._arduino.subscribe_digital(self.dev, self.pin, self)

    def cleanup(self):
        self._arduino.unsubscribe_digital(self.dev, self.pin)
